import json
from typing import Dict, Iterable, List, Optional, Tuple, Union
from urllib.parse import ParseResult, urlparse

import numpy as np
import pandas as pd
import pkg_resources
from intake import DataSource, Schema
from intake.catalog.local import LocalCatalogEntry


class DalSource(DataSource):
    """
    DalSource abstracts a dataset over heterogeneous storage systems.
    DalSource follows the pattern of Intake AliasSource, which in the end
    refers to another driver.

    The ending driver must accept an "urlpath" argument because the URL is
    parsed as such:
     - scheme is the Intake driver name.
     - path becomes the driver "urlpath" argument.

    user_events:
      driver: dal
      args:
        default: 'local'
        storage:
          local: 'csv://{{ CATALOG_DIR }}/data/user_events.csv'
          serving: 'in-memory-kv://foo'
          batch: 'parquet://{{ CATALOG_DIR }}/data/user_events.parquet'
    """

    container = "dataframe"
    name = "dal"
    version = pkg_resources.get_distribution("intake-dal").version

    def __init__(self, storage, default, storage_mode=None, metadata=None, **kwargs):
        """
        Parameters
        ----------
        path: str
            Location of the file to parse (can be remote)
        reload : bool
            Whether to watch the source file for changes; make False if you want
            an editable Catalog
        """
        super(DalSource, self).__init__(metadata)
        self.storage = storage
        self.default = default
        self.storage_mode = storage_mode
        self.kwargs = kwargs
        self.metadata = metadata
        self.source = None
        self._canonical_name = None  # _get_schema() sets this
        self._avro_schema = None  # _get_schema() sets this
        self._dtypes = None  # _get_schema() sets this

    def _get_source(self):
        if self.catalog_object is None:
            raise ValueError("DalSource cannot be used outside a catalog")
        if self.source is None:
            self._get_schema()
            self.source = self._instantiate_source()
            self.metadata = self.source.metadata.copy()
            self.container = self.source.container
            self.partition_access = self.source.partition_access
            self.description = self.source.description
            self.datashape = self.source.datashape

    def _get_schema(self) -> Schema:
        if self._canonical_name is None:

            self._canonical_name = _get_dal_canonical_name(self)
            # TODO(talebz): Getting avro schema should be promoted to Intake
            avro_schema = _get_avro(self, self._canonical_name)
            if avro_schema:
                self._avro_schema = avro_schema
                self._schema_dtypes = _avro_to_dtype(self._avro_schema)
                self._dtypes = {k: str(v) for (k, v) in self._schema_dtypes.items()}

        return Schema(
            datashape=None,
            dtype=self._dtypes,
            shape=(None, len(self._dtypes) if self._dtypes else None),
            npartitions=1,  # This data is not partitioned, so there is only one partition
            extra_metadata={
                "canonical_name": self._canonical_name,
                "storage_mode": self.storage_mode,
                "avro_schema": self._avro_schema,
            },
        )

    @staticmethod
    def parse_storage_mode_url(mode_url: str) -> Tuple[ParseResult, str]:
        """
        :param mode_url:
            scheme is the Intake driver name.
            examples:
              - "csv://{{ CATALOG_DIR }}/data/user_events.csv"
              - "dal-online://http://0.0.0.0:9166#zpid"
              - "hive://user_events_dal_catalog2;userid={{userid}}"
        :return:
            parse_result and url_path without the scheme, this is sent to the DataSource
            examples:
              - "{{ CATALOG_DIR }}/data/user_events.csv"
              - "http://0.0.0.0:9166#zpid"
              - "user_events_dal_catalog2;userid={{userid}}"
        """
        parse_result: ParseResult = urlparse(mode_url)
        fragment = "" if parse_result.fragment == "" else f"#{parse_result.fragment}"
        query = "" if parse_result.query == "" else f"?{parse_result.query}"

        url_path = f"{parse_result.netloc}{parse_result.path}{parse_result.params}{query}{fragment}"
        return parse_result, url_path

    def _instantiate_source(self):
        """ Driving method of this class. """
        mode = self.storage[self.storage_mode if self.storage_mode else self.default]

        args = {}
        mode_url = mode
        if isinstance(mode, dict):
            mode_url = mode["url"]
            args = mode.get("args", {})

        parse_result, url_path = self.parse_storage_mode_url(mode_url)
        desc = self.catalog_object[self.name].describe()

        if parse_result.scheme == "parquet":
            # https://github.com/dask/dask/issues/5272: Dask parquet metadata w/ ~2k files very slow
            if "gather_statistics" not in args:
                args["gather_statistics"] = False

            if "engine" not in args:
                args["engine"] = "pyarrow"

        entry = LocalCatalogEntry(
            name=desc["name"],
            description=desc["description"],
            driver=parse_result.scheme,
            args={"urlpath": url_path, **args},
            parameters=self.catalog_object[self.name]._user_parameters,
            catalog=self.cat,
        )

        params = {
            "canonical_name": self._canonical_name,
            "storage_mode": self.storage_mode,
            "avro_schema": self._avro_schema,
            "dtypes": self._dtypes,
        }

        source = entry.get(metadata=self.metadata, **self.kwargs)
        # source = entry.get(metadata=self.metadata, **{**self.kwargs, **params})

        source.metadata["url_path"] = url_path
        source.metadata = {**source.metadata, **params}

        return source

    def discover(self):
        self._get_source()
        return self.source.discover()

    def read(self):
        self._get_source()
        return self.source.read()

    def read_partition(self, i):
        self._get_source()
        return self.source.read_partition(i)

    def read_chunked(self):
        self._get_source()
        return self.source.read_chunked()

    # TODO(talebz): This should also be within Intake but without DataFrame type!
    def write(self, df: pd.DataFrame):
        self._get_source()
        return self.source.write(df)

    def to_spark(self):
        self._get_source()
        return self.source.to_spark()

    def to_dask(self):
        self._get_source()
        return self.source.to_dask()

    @property
    def avro_schema(self) -> Dict:
        self._get_source()
        return self._avro_schema

    @property
    def canonical_name(self) -> Dict:
        self._get_source()
        return self._canonical_name


def _get_dal_canonical_name(source: DataSource) -> str:
    def helper(source: DataSource) -> List[str]:
        if source.cat is None:
            return []  # the parent catalog is not part of the canonical name
        elif source.cat:
            return helper(source.cat) + [source.name]

    return ".".join(helper(source))


def _get_avro(source: DataSource, canonical_name: str) -> Optional[Dict]:
    data_schema_entry = _get_metadata_schema(source)
    if data_schema_entry is None:
        return None

    if "kafka_schema_registry" in data_schema_entry:
        # TODO(talebz): check data_schema_entry for kafka_schema_registry.  If exists then query Kafka Schema Registry
        raise NotImplementedError(
            "kafka_schema_registry integration not yet supported.  "
            f"Please put schema as {canonical_name}: > JSON of avro schema"
        )

    if canonical_name in data_schema_entry:
        return json.loads(data_schema_entry[canonical_name])
    else:
        return None


def _get_metadata_schema(source: DataSource) -> Dict:
    if "data_schema" in source.metadata:
        return source.metadata["data_schema"]
    elif source.cat:
        return _get_metadata_schema(source.cat)


# TODO(talebz): ensure this is comprehensive with unit tests!
def _avro_to_dtype(schema: Dict) -> Dict:
    field_schemas = {f["name"]: f["type"] for f in schema["fields"]}
    avro_type_to_dtype = {
        tuple(sorted(["type", "long", "logicalType", "timestamp-millis"])): np.dtype("datetime64"),
        tuple(sorted(["type", "long", "logicalType", "timestamp-micros"])): np.dtype("datetime64"),
        tuple(sorted(["null", "int"])): np.dtype("int32"),
        tuple(sorted(["null", "long"])): np.dtype("int32"),
        tuple(sorted(["type", "int", "unsigned", "True"])): np.dtype("uint32"),
        tuple(sorted(["type", "long", "unsigned", "True"])): np.dtype("int64"),
        tuple(["long"]): np.dtype("int64"),
        tuple(["int"]): np.dtype("int32"),
        tuple(["float"]): np.dtype("float32"),
        tuple(["double"]): np.dtype("float64"),
        tuple(["boolean"]): np.dtype("bool"),
        tuple(["string"]): np.dtype("object"),
    }

    def to_lookup(avro_type: Union[str, list]) -> tuple:
        if isinstance(avro_type, str):
            return tuple([avro_type])
        elif isinstance(avro_type, list):
            return tuple(sorted(_flatten(avro_type)))

    ret = {}
    for (k, v) in field_schemas.items():
        lookup = to_lookup(v)
        if lookup in avro_type_to_dtype:
            ret[k] = avro_type_to_dtype[lookup]
        elif "null" in lookup:
            list_lookup = list(lookup)
            list_lookup.remove("null")
            new_lookup = tuple(list_lookup)
            if new_lookup in avro_type_to_dtype:
                ret[k] = avro_type_to_dtype[new_lookup]
            else:
                raise ValueError(f"{lookup} to pandas type not supported in {schema}")
        else:
            raise ValueError(f"{lookup} to pandas type not supported in {schema}")

    return ret


def _flatten(ls: Iterable) -> Iterable:
    def iter_ls():
        if isinstance(ls, dict):
            return ls.items()
        else:
            return ls

    for i in iter_ls():
        if isinstance(i, Iterable) and not isinstance(i, str):
            for sub_collection in _flatten(i):
                yield sub_collection
        else:
            yield i



# to avoid circular dependency: importing the package imports intake, which tries
# to discover and import packages that define drivers. To avoid this pitfall,
# just ensure that ``intake`` is imported first thing in your package's  ``__init__.py``
import intake  # noqa: F401



import datetime
import json
from typing import Dict, List
from unittest import mock
from unittest.mock import MagicMock

import pandas as pd
import pytest
from pandas.util.testing import assert_frame_equal

from intake_dal.dal_catalog import DalCatalog
from intake_dal.dal_online import (
    DalOnlineSource,
    deserialize_avro_str_to_pandas,
    serialize_panda_df_to_str,
)


@pytest.fixture
def user_events_df():
    return pd.DataFrame(
        {
            "userid": [100, 101],
            "home_id": [3, 4],
            "action": ["click", "click"],
            "timestamp": [datetime.datetime(2012, 5, 1, 0, 0), datetime.datetime(2012, 5, 2, 0, 0)],
        }
    )


@pytest.fixture
def user_events_json():
    return [
        {
            "userid": 100,
            "home_id": 3,
            "action": "click",
            "timestamp": {"format": "DATETIME", "time": "2012-05-01 00:00:00.000000"},
        },
        {
            "userid": 101,
            "home_id": 4,
            "action": "click",
            "timestamp": {"format": "DATETIME", "time": "2012-05-02 00:00:00.000000"},
        },
    ]


@mock.patch("intake_dal.dal_online._http_put_avro_data_set")
@mock.patch("intake_dal.dal_online._http_get_avro_data_set")
def test_dal_online_write_read(
    mock_get: MagicMock,
    mock_put: MagicMock,
    serving_cat: DalCatalog,
    user_events_df: pd.DataFrame,
    user_events_json: List[Dict],
):
    canonical_name = "entity.user.user_events"
    avro_str = serialize_panda_df_to_str(
        user_events_df, schema=json.loads(serving_cat.metadata["data_schema"][canonical_name])
    )
    mock_get.return_value = user_events_json
    mock_put.return_value = 200

    serving_cat.entity.user.user_events(storage_mode="serving").write(user_events_df)

    assert_frame_equal(user_events_df, serving_cat.entity.user.user_events(key=100).read(), check_dtype=False)
    mock_get.assert_called()
    mock_put.assert_called()

    assert_frame_equal(
        user_events_df,
        deserialize_avro_str_to_pandas(mock_put.call_args[0][1]["avro_rows"]),
        check_dtype=False,
    )

    assert avro_str != mock_put.call_args[0][1]["avro_rows"]  # not sure why!

    # todo: Why is this not idempotent?
    assert serialize_panda_df_to_str(
        user_events_df, schema=json.loads(serving_cat.metadata["data_schema"][canonical_name])
    ) != serialize_panda_df_to_str(
        user_events_df, schema=json.loads(serving_cat.metadata["data_schema"][canonical_name])
    )


def test_dal_write_parallelism(serving_cat: DalCatalog):
    assert (
        serving_cat["entity.user.user_events"].discover()["metadata"][DalOnlineSource.name][
            "write_parallelism"
        ]
        == 2
    )
    assert (
        serving_cat.entity.user.user_events.discover()["metadata"][DalOnlineSource.name]["write_parallelism"]
        == 2
    )


@mock.patch("intake_dal.dal_online._http_put_avro_data_set")
def test_post_in_chunks(mock_put: MagicMock, serving_cat: DalCatalog, user_events_df: pd.DataFrame):
    mock_put.return_value = 200

    ds: DalOnlineSource = serving_cat["entity.user.user_events"]
    ds.metadata[DalOnlineSource.name]["write_chunk_size"] = 1
    ret = ds.write(user_events_df)

    assert len(mock_put.call_args_list) == 2
    assert len(ret) == 2



import base64
import io
import time
import urllib.parse
from datetime import datetime
from http import HTTPStatus
from typing import Callable, Dict, List, Tuple
from urllib.parse import ParseResult, urldefrag, urlparse  # noqa: F401

import numpy as np
import pandas as pd
import pandavro
import pkg_resources
import requests
from intake import DataSource, Schema


class DalOnlineSource(DataSource):
    """
    Interfaces with Online FS
    This Intake source is to be used in conjunction with intake-dal
    """

    container = "dataframe"
    partition_access = False
    name = "dal-online"
    version = pkg_resources.get_distribution("intake-dal").version

    DATE_TIME_FORMAT = "%Y-%m-%d %H:%M:%S.%f"

    def __init__(self, urlpath, key=None, storage_options=None, metadata=None):
        """
            Fetches rows from the Online Feature Store (FS).
        Args:
            urlpath: the host path, must include primary_key name as a fragment.
            key: The primary_key value.
            storage_options: optional storage options to send along
            metadata: Used by Intake
        """

        self._urlpath = urlpath
        parse_result = urlparse(urlpath)  # type: ParseResult
        self._url, _ = urldefrag(urlpath)
        self._key_name = parse_result.fragment
        if self._key_name == "":
            raise ValueError(f"key_name expected in URL fragment of {urlpath}")

        self._canonical_name = None  # set in _get_schema()
        self._key_value = key
        super().__init__(metadata=metadata)

    def write(self, df: pd.DataFrame) -> List[Tuple[float, float]]:
        self._get_schema()

        def post_lambda(avro_str: str) -> int:
            return _http_put_avro_data_set(
                self._url,
                {"data_set_name": self._canonical_name, "key_value": self._key_name, "avro_rows": avro_str},
            )

        def get_metadata(key: str, default):
            if DalOnlineSource.name in self.metadata:
                return self.metadata[DalOnlineSource.name].get(key, default)
            else:
                return default

        # get settings with defaults
        write_chunk_size = get_metadata("write_chunk_size", default=1000)
        write_delay_between_chunks_milliseconds = get_metadata(
            "write_delay_between_chunks_milliseconds", default=50
        )
        return _post_in_chunks(
            df, self._avro_schema, post_lambda, write_chunk_size, write_delay_between_chunks_milliseconds
        )

    def _get_partition(self, _) -> pd.DataFrame:
        self._get_schema()
        data = _http_get_avro_data_set(self._url, self._canonical_name, self._key_value)
        for row in data:
            for key, field in row.items():
                if isinstance(field, dict) and "format" in field:
                    row[key] = datetime.strptime(field["time"], self.DATE_TIME_FORMAT)
        return pd.DataFrame(data)

    def _close(self):
        pass

    def _get_schema(self) -> Schema:
        if self._canonical_name is None:
            self._canonical_name = self.metadata["canonical_name"]
            self._avro_schema = self.metadata["avro_schema"]
            self._dtypes = self.metadata["dtypes"]
            self._storage_mode = self.metadata["storage_mode"]

        return Schema(
            datashape=None,
            dtype=self._dtypes,
            shape=(None, len(self._dtypes)),
            npartitions=1,  # This data is not partitioned, so there is only one partition
            extra_metadata={
                "canonical_name": self._canonical_name,
                "storage_mode": self._storage_mode,
                "avro_schema": self._avro_schema,
            },
        )


AVRO_DATA_SETS_PATH = "avro-data-sets"


def _post_in_chunks(
    df: pd.DataFrame,
    avro_schema: Dict,
    post_lambda: Callable[[str], int],
    write_chunk_size: int,
    write_delay_between_chunks_milliseconds: int,
) -> List[Tuple[float, float]]:
    """
    :param df: DataFrame to post
    :param post_lambda: Lambda to pust the avro to and it returns the status code.
    :return: list of durations of how long it took to (serialize to Avro, run post_lambda)
    """
    times = []
    number_of_chunks = np.math.ceil(len(df) / write_chunk_size)
    for i, chunk in enumerate(np.array_split(df, number_of_chunks)):
        if i != 0:
            time.sleep(write_delay_between_chunks_milliseconds / 1000)  # sleep takes seconds

        avro_begin_time = time.time()
        avro_str = serialize_panda_df_to_str(chunk, avro_schema)
        avro_time = time.time() - avro_begin_time

        post_begin_time = time.time()
        post_lambda(avro_str)
        times.append((avro_time, time.time() - post_begin_time))
    return times


def _http_get_avro_data_set(url: str, canonical_name: str, key_value: str) -> List[Dict]:
    response = requests.get(urllib.parse.urljoin(url, f"{AVRO_DATA_SETS_PATH}/{canonical_name}/{key_value}"))
    if response.status_code != HTTPStatus.OK.value:
        raise Exception(f"url={response.url} code={response.status_code}: {response.text}")
    return response.json()["data"]


def _http_put_avro_data_set(url: str, json: Dict) -> int:
    response = requests.put(urllib.parse.urljoin(url, f"{AVRO_DATA_SETS_PATH}/"), json=json)
    if response.status_code != HTTPStatus.OK.value:
        raise Exception(f"url={response.url} code={response.status_code}: {response.text}")
    return response.status_code


def serialize_panda_df_to_str(df: pd.DataFrame, schema: Dict) -> str:
    with io.BytesIO() as bytes_io:
        # else we get: ValueError: NaTType does not support timestamp
        # it's really a pandavro issue, see https://github.com/fastavro/fastavro/issues/313
        # TODO(talebz): Create a Pandavro issue for this!
        df = df.replace({np.nan: None})
        pandavro.to_avro(bytes_io, df, schema=schema)
        bytes_io.seek(0)
        return base64.b64encode(bytes_io.read()).decode("utf-8")


def deserialize_avro_str_to_pandas(avro_str: str, schema: dict = None) -> pd.DataFrame:
    return pandavro.from_avro(io.BytesIO(base64.b64decode(avro_str)), schema)



import functools

import pkg_resources
import yaml
from intake import Catalog
from intake.utils import yaml_load
from intake_nested_yaml_catalog.nested_yaml_catalog import (
    NestedYAMLFileCatalog,
)

from intake_dal.dal_source import DalSource


class DalCatalog(NestedYAMLFileCatalog):
    """
    DalCatalog combines the functionality of a nested hierarchical catalog
    along with a the "dal" DataSource.
    """

    name = "dal_cat"
    version = pkg_resources.get_distribution("intake-dal").version

    def __init__(self, path, storage_mode=None, autoreload=True, **kwargs):
        """
        Parameters
        ----------
        path: str
            Location of the file to parse (can be remote)
        reload : bool
            Whether to watch the source file for changes; make False if you want
            an editable Catalog
        storage_mode: str
            The dal default storage mode override for this instantiation of the
            catalog.

        Example catalog:
          sources:
            user_events:
              driver: dal
              args:
                default: 'local'
                storage:
                  local: 'csv://{{ CATALOG_DIR }}/data/user_events.csv'
                  serving: 'in-memory-kv://foo'
                  batch: 'parquet://{{ CATALOG_DIR }}/data/user_events.parquet'

        Following overrides the default from 'local' to 'serving'.
        >>> cat = DalCatalog(path, storage_mode="serving")
        >>> df = cat.user_events.read()
        """
        self.storage_mode = storage_mode
        super(DalCatalog, self).__init__(path, autoreload, **kwargs)

    def __getitem__(self, key):
        # TODO(Taleb Zeghmi): Remove once https://github.com/zillow/intake-nested-yaml-catalog/issues/6 is resolved
        if len(key.split(".")) > 1:
            return self._construct_dataset(key, self)
        else:
            ret = super().__getitem__(key)
            return ret

    def parse(self, text):
        data = yaml_load(text)

        # modify sources default storage mode
        self._set_dal_default_storage_mode(data)
        transformed_text = yaml.dump(data, default_flow_style=False)

        # Reuse default NestedYAMLFileCatalog YAML parser
        # parse() does the heavy lifting of populating the catalog
        super().parse(transformed_text)

    def _set_dal_default_storage_mode(self, data):
        """
        Traverses the catalog to set all default dal source
        storage modes to self.storage_mode
        """
        if self.storage_mode:
            for k, v in data.items():
                if isinstance(v, dict) and v.get("driver", None) == "dal":
                    v["args"]["default"] = self.storage_mode
                elif isinstance(v, dict):
                    self._set_dal_default_storage_mode(v)

    @staticmethod
    def _construct_dataset(canonical_name: str, catalog: Catalog) -> DalSource:
        catalog_entity = functools.reduce(lambda acc, x: acc[x], canonical_name.split("."), catalog)
        return catalog_entity
